import pickle
import time
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import Dataset
from torchvision import models
from tqdm import tqdm
import matplotlib.pyplot as plt
import matplotlib as mpl
from unet import UNet
import torchvision
import shutil
import os  # optional: for extracting basename / creating new filepath
import time  # optional: for appending time string to copied script

from additionals import *

# Specify data and output folders
# pth_data = '/projects/in5400/in5400_2021_lab_segmentation_data'
pth_data = '/nr/samba/user/arej/PycharmProjects/in5400_2021_lab_segmentation'
# pth_work = None #  [optional] '/path-to-somewhere-you-can-store-intermediate-results/'
pth_work = '/nr/samba/user/arej/PycharmProjects/in5400_2021_lab_segmentation/wdir_20210319'

if pth_work is not None:
    # Make a copy of the current script
    copied_script_name = time.strftime("%Y%m%d_%H%M") + '_' + os.path.basename(__file__)  # generate filename with timestring
    shutil.copy(__file__, os.path.join(pth_work, copied_script_name))  # copy script
    print('Wrote file: {}'.format(copied_script_name))

# Read list of files used to train
with open(os.path.join(pth_data, 'annotations/trainval.txt')) as f:
    lines = f.read().splitlines()
im_list_train = [line.split(' ')[0]+'.jpg' for line in lines]

# Read list of files used for validation
with open(os.path.join(pth_data, 'annotations/test.txt')) as f:
    lines = f.read().splitlines()
im_list_val = [line.split(' ')[0]+'.jpg' for line in lines]

mpl.use('Agg')

# Create data loaders for both the train and validation set
pth_images = os.path.join(pth_data, 'images/')
pth_gt = os.path.join(pth_data, 'annotations/trimaps/')
datasets = {'train': PetDataset(im_list_train, pth_images, pth_gt), 'val': PetDataset(im_list_val, pth_images, pth_gt)}
dataloaders = {split: torch.utils.data.DataLoader(datasets[split], batch_size=3, shuffle=True, pin_memory=True) for split in ('train', 'val')}

if 0:
    # ### Visualize an example image and ground-truth mask from the training data
    ind = random.choice(range(len(datasets['train'])))
    im, im_gt = datasets['train'][ind]

    im = np.transpose(im.detach().numpy(), [1, 2, 0])
    plt.imshow(im)
    plt.show()
    plt.imshow(im_gt)
    plt.show()

device = 'cuda'  # 'cuda'/'cpu'

# model = UNet(chs=(3, 64, 128, 256), num_class=3)
model = UNet(chs=(3, 64, 128, 256, 512, 1024, 2048), num_class=3)

# Verify that the model produces an output layer having the correct shape
out = model(torch.rand((1, 3, 224, 224))).detach()  # Run a single dummy-image through the model
assert out.shape == (1, 3, 224, 224), "The output of the model for a 224 x 224 input image should be num_class x 224 x 224."

model.to(device=device)  # Move the model onto GPU if so opted for

# Specify the loss function
class_weights = torch.tensor([1, 1, 2], dtype=torch.float)
criterion = nn.CrossEntropyLoss(weight=class_weights).to(device=device)

# Specify the optimizer and the learning-rate scheduler
optimizer = optim.SGD(model.parameters(), lr=1e-2, momentum=0.9, weight_decay=1e-5)
scheduler = optim.lr_scheduler.CyclicLR(optimizer, base_lr=0, max_lr=1e-2, step_size_up=500, step_size_down=1000, base_momentum=0.9, max_momentum=0.9, mode='triangular2')  # Move from zero to 1e-2 lr in 500 iterations, then back down in 1000 iterations; iterate while halving the maximum amplitude

total_epochs = 150  # Specify for how long we are prepared to train the model

epoch_loss = {'train': [], 'val': []}
epoch_accuracy = {'train': [], 'val': []}
epoch_mean_accuracy = {'train': [], 'val': []}
epoch_mean_IoU = {'train': [], 'val': []}
epoch_freq_IoU = {'train': [], 'val': []}
epoch_metrics_and_more = {'train': [], 'val': []}

# Train the model by looping through all the images 'total_epochs' times
for epoch in range(total_epochs):
    for phase in ('train', 'val'):
        is_training = phase == 'train'
        model.train(is_training)
        torch.set_grad_enabled(is_training)
        torch.cuda.empty_cache()

        start_time = time.time()

        total_loss = 0
        total_accuracy = 0
        total_mean_accuracy = 0
        total_mean_IoU = 0
        total_freq_IoU = 0
        total_images = 0  # XXX NOTE: misleading!  use total_iterations or whatever

        for ix, (ims, ims_gt) in enumerate(tqdm(dataloaders[phase])):

            if ix > 100:   # Cut the epochs short, a mere 100 iterations
                break

            # Put the images and targets on GPU if so opted for
            ims = ims.to(device=device)
            targets = ims_gt.type(torch.long).to(device=device)

            # Run images through the model, producing a num_class x 224 x 224 output for each
            outputs = model(ims)

            # During the first iteration, write a sample image, output and target
            if ix == 0:
                write_sample(ims, outputs, targets, pth_work, epoch, phase)

            # Calculate the loss
            loss = criterion(outputs, targets)

            # Accumulate losses for later display
            total_loss += loss.item()
            total_images += ims.shape[0]

            # Calculate accuracy, mean accuracy, mean IoU and freq-weighted IoU
            y = np.argmax(outputs.detach().cpu().numpy(), 1)
            total_accuracy += np.sum(y==targets.cpu().numpy()) / np.prod(y.shape) * ims.shape[0]  # Number of correct pixels divided by the total amount of them, times batch size
            macc = 0
            for i in range(0, 3):
                xxx = targets.cpu().numpy()==i
                t = targets.cpu().numpy()
                macc += np.sum(y[xxx] == t[xxx]) / np.sum(xxx)
            total_mean_accuracy += macc / 3 * ims.shape[0]  # note: later, when total_images is replaced, remove this im.shape[0] factor

            mIoU = 0
            for i in range(0, 3):
                t = targets.cpu().numpy()
                xxx = t == i
                yyy = y == i
                intersection = np.sum(np.abs(xxx * yyy))
                union = np.sum(xxx) + np.sum(yyy) - intersection
                iou = intersection / union
                mIoU += iou
            total_mean_IoU += mIoU / 3 * ims.shape[0]  # note: later, when total_images is replaced, remove this im.shape[0] factor

            fwIoU = 0
            w_tot = 0
            for i in range(0, 3):
                t = targets.cpu().numpy()
                xxx = t == i
                yyy = y == i
                intersection = np.sum(np.abs(xxx * yyy))
                union = np.sum(xxx) + np.sum(yyy) - intersection
                iou = intersection / union
                w = np.count_nonzero(xxx)
                fwIoU += iou * w
                w_tot += w
            total_freq_IoU += fwIoU / w_tot * ims.shape[0]  # note: later, when total_images is replaced, remove this im.shape[0] factor

            if is_training:
                # Update the weights by doing back-propagation
                # [Note that by doing this last, the CPU can work on loading images for the next iteration while the GPU handles this task]
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                scheduler.step()

            del loss, outputs  # Make sure we allow freeing these on GPU to save on memory

        elapsed_time = time.time() - start_time
        if phase == 'train':
            print('[{}]\tEpoch: {}/{}\t loss: {:.5f}\tlr: {:.7f}\taccuracy: {:.5f}\tmAcc: {:.5f}\tmIoU: {:.5f}\tfwIoU: {:.5f}'.format(phase, epoch+1, total_epochs, total_loss/total_images, get_lr(optimizer), total_accuracy/total_images, total_mean_accuracy/total_images, total_mean_IoU/total_images, total_freq_IoU/total_images))
        else:
            print('[{}]\tEpoch: {}/{}\t loss: {:.5f}\tlr: {:.7f}\taccuracy: {:.5f}\tmAcc: {:.5f}\tmIoU: {:.5f}\tfwIoU: {:.5f}'.format(phase, epoch+1, total_epochs, total_loss/total_images, get_lr(optimizer), total_accuracy/total_images, total_mean_accuracy/total_images, total_mean_IoU/total_images, total_freq_IoU/total_images))

        epoch_loss[phase].append(total_loss/total_images)
        epoch_accuracy[phase].append(total_accuracy / total_images)
        epoch_mean_accuracy[phase].append(total_mean_accuracy/total_images)
        epoch_mean_IoU[phase].append(total_mean_IoU/total_images)
        epoch_freq_IoU[phase].append(total_freq_IoU/total_images)

        if pth_work is not None:
            with open(os.path.join(pth_work, 'xxx_{}.txt'.format(phase)), 'w') as fh:
                fh.writelines("{:.5f}\n".format(loss) for loss in epoch_loss[phase])

    if pth_work is not None:
        # Plot train / val losses and write it to file
        plt.clf()
        plt.plot(epoch_loss['train'], label='train')
        plt.plot(epoch_loss['val'], label='val')
        plt.xlabel('epoch')
        plt.ylabel('loss')
        plt.legend()
        plt.grid(True)
        fname1 = os.path.join(pth_work, 'fig_train_val_loss.png')
        plt.savefig(fname1)

        # Plot train / val metrics and write it to file
        plt.clf()
        plt.plot(epoch_accuracy['train'], label='acc, train')
        plt.plot(epoch_accuracy['val'], label='acc, val')
        plt.plot(epoch_mean_accuracy['train'], label='mAcc, train')
        plt.plot(epoch_mean_accuracy['val'], label='mAcc, val')
        plt.plot(epoch_mean_IoU['train'], label='mIoU, train')
        plt.plot(epoch_mean_IoU['val'], label='mIoU, val')
        plt.plot(epoch_freq_IoU['train'], label='fwIoU, train')
        plt.plot(epoch_freq_IoU['val'], label='fwIoU, val')
        plt.xlabel('epoch')
        # plt.ylabel('loss')
        plt.legend()
        plt.grid(True)
        fname2 = os.path.join(pth_work, 'fig_train_val_acc_mAcc_mIoU_fwIoU_{}.png'.format(os.path.basename(pth_work)))
        plt.savefig(fname2)
        print('Wrote {} and {}'.format(fname1, fname2))

